"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseSpdxExpressionWithDetails = exports.parse = void 0;
var strict_parser_1 = require("./strict_parser");
var liberal_parser_1 = require("./liberal_parser");
var licenses_1 = require("../licenses");
var validator_1 = __importDefault(require("../validator"));
/**
 * Clean up the given SPDX expression before processing it with the grammar.
 *
 * Currently, the clean-up is limited to:
 * 1. trimming the surrounding whitespace, and
 * 2. compressing consecutive whitespace characters into a single space.
 *
 * @param input SPDX expression as a string
 * @param strictSyntax If {true}, don't clean up spec-violating details
 * @returns A slightly cleaned up SPDX expression as a string
 */
var prepareInput = function (input, strictSyntax) {
    var cleanedInput = input
        .trim() // remove surrounding whitespace
        .replace(/\s+/, ' '); // compress consecutive whitespaces to a single space
    if (!strictSyntax) {
        cleanedInput = cleanedInput
            .replace(/(?<!\s)\s+and\s+/i, ' AND ') // fix lowercase keywords
            .replace(/(?<!\s)\s+or\s+/i, ' OR ') // fix lowercase keywords
            .replace(/(?<!\s)\s+with\s+/i, ' WITH ') // fix lowercase keywords
            .replace(/\s[wW]\/\s?/, ' WITH '); // expand "w/" shorthand for "WITH"
    }
    return cleanedInput;
};
var compileFullSpdxParseResult = function (input, parserResult) {
    var _a;
    return { input: input, ast: (_a = parserResult.parse) === null || _a === void 0 ? void 0 : _a.ast, error: parserResult.error, expression: parserResult.expression };
};
var buildErrorMessage = function (input, strictSyntax) {
    return "".concat(strictSyntax ? 'Strict' : 'Liberal', " parsing for ").concat(JSON.stringify(input), " failed");
};
/**
 * Parse an SPDX expression into a structured object representation.
 *
 * @param input SPDX expression as a string to be parsed.
 * @returns {ParsedSpdxExpression} A structured object describing the given SPDX expression
 *          or throws an `Error` if parsing failed.
 */
function parse(input, parseOptions) {
    var strictSyntax = !!(parseOptions === null || parseOptions === void 0 ? void 0 : parseOptions.strictSyntax);
    var upgradeGPLVariants = !!(parseOptions === null || parseOptions === void 0 ? void 0 : parseOptions.upgradeGPLVariants);
    var data = parseSpdxExpressionWithDetails(input, strictSyntax, upgradeGPLVariants);
    if (data.error) {
        throw new Error([buildErrorMessage(input, strictSyntax), data.error].join(': '));
    }
    else if (data.expression === undefined) {
        /* istanbul ignore next */ throw new Error(buildErrorMessage(input, strictSyntax));
    }
    return data.expression;
}
exports.parse = parse;
/**
 * Preprocess the original input for the liberal parser, detecting ~full license name matches
 * that would syntactically fail parsing (e.g. due to parenthesis or keywords such as "and").
 *
 * @param input The input to "clean up" before lexical parsing
 * @returns The preprocessed string ready to be fed to the parser
 */
var prepareLiberalInput = function (input) {
    // List of (hand-crafted) mutations to apply, replacing license names with a matching identifier
    var mutations = [
        // CDDL
        function (input) {
            var pattern = /Common Development and Distribution License( \(CDDL\))?( (v|version )?(1\.[10]))?/im;
            var match = input.match(pattern);
            if (match) {
                var replacement = !!(match[4]) ? 'CDDL-$4' : 'CDDL-1.1'; // default to the latest version, i.e. 1.1
                return input.replace(pattern, replacement);
            }
            return input;
        },
        // FSF Unlimited License (With License Retention    and Warranty Disclaimer)
        function (input) {
            var pattern = /FSF\s+Unlimited\s+License\s+\(With\s+License\s+Retention\s+and\s+Warranty\s+Disclaimer\)/im;
            var match = input.match(pattern);
            if (match) {
                return input.replace(pattern, 'FSFULLRWD');
            }
            return input;
        }
    ];
    // Automated, exact-name based mutations to apply, replacing license names with a matching identifier
    var relevantLicenses = licenses_1.licenses.licenses.filter(function (license) { return license.name.includes('(') || license.name.match(/ and /i); });
    relevantLicenses.forEach(function (license) {
        mutations.push(function (input) {
            return input.replace(license.name, license.licenseId);
        });
    });
    return mutations.reduce(function (result, mutation) { return mutation(result); }, input);
};
/**
 * Parse an SPDX expression into a structured object representation along with additional
 * metadata such as the underlying AST tree used and any errors in case parsing failed.
 *
 * @param input SPDX expression as a string to be parsed.
 * @returns {FullSpdxParseResult}
 */
function parseSpdxExpressionWithDetails(input, strictSyntax, upgradeGPLVariants) {
    var _a;
    if (strictSyntax === void 0) { strictSyntax = false; }
    if (upgradeGPLVariants === void 0) { upgradeGPLVariants = false; }
    // Apply general clean up on the raw input string
    var preparedInput = prepareInput(input, strictSyntax);
    // Always try to parse with the strict parser first in order to
    // minimize risk of unwanted "corrections"
    var notCorrected = (0, strict_parser_1.parse)(preparedInput, upgradeGPLVariants);
    if (!!notCorrected.expression && !strictSyntax) {
        notCorrected.expression = correctDashSeparatedLicenseIds(notCorrected.expression, upgradeGPLVariants);
    }
    var strictResult = compileFullSpdxParseResult(preparedInput, notCorrected);
    // If strict parsing failed, attempt to apply corrections if allowed
    if (strictResult.error && !strictSyntax) {
        var liberallyPreparedInput = prepareLiberalInput(preparedInput);
        var corrected = (0, liberal_parser_1.parse)(liberallyPreparedInput, upgradeGPLVariants);
        if (corrected.expression) {
            corrected.expression = correctDashSeparatedLicenseIds(corrected.expression, upgradeGPLVariants);
        }
        var liberalResult = compileFullSpdxParseResult(liberallyPreparedInput, corrected);
        // If the license identifier happens to be an exact match of a license name...
        var nameBasedMatch = (0, licenses_1.findNameBasedMatch)(((_a = liberalResult.expression) === null || _a === void 0 ? void 0 : _a.license) || preparedInput, upgradeGPLVariants);
        if (nameBasedMatch !== undefined) {
            return parseSpdxExpressionWithDetails(nameBasedMatch.licenseId, true, upgradeGPLVariants);
        }
        // We'll use the results of liberal parsing only if it succeeds. In case
        // of failure, we'll return the errors from strict parser for maximum
        // awareness of deviations from the SPDX specification
        if (!liberalResult.error) {
            return liberalResult;
        }
        // If the value looks like "Mozilla Public License 2.0 (MPL 2.0)", and is short enough
        // to look like a license name, try to parse the identifier in the parenthesis.
        // If that fails, we'll try to see if the parenthesized text is an exact match against
        // a license name in the SPDX data.
        var potentialMatchForParenthesizedPattern = preparedInput === null || preparedInput === void 0 ? void 0 : preparedInput.match(/^(.+?)\s\((.+?)\)$/);
        var pretext = potentialMatchForParenthesizedPattern === null || potentialMatchForParenthesizedPattern === void 0 ? void 0 : potentialMatchForParenthesizedPattern[1];
        var potentialIdentifier = potentialMatchForParenthesizedPattern === null || potentialMatchForParenthesizedPattern === void 0 ? void 0 : potentialMatchForParenthesizedPattern[2];
        if (pretext && potentialIdentifier) {
            // check if the pretext is within threshold length
            var thresholdLength = 100;
            if (pretext.length < thresholdLength) {
                // check if it's an SPDX identifier
                var candidateResult = parseSpdxExpressionWithDetails(potentialIdentifier, strictSyntax, upgradeGPLVariants);
                if (!candidateResult.error && candidateResult.expression && (0, validator_1.default)(candidateResult.input).valid) {
                    return candidateResult;
                }
            }
            // check for a name-based match
            var nameBasedMatch_1 = (0, licenses_1.findNameBasedMatch)(potentialIdentifier, upgradeGPLVariants);
            if (nameBasedMatch_1) {
                return parseSpdxExpressionWithDetails(nameBasedMatch_1.licenseId, strictSyntax, upgradeGPLVariants);
            }
        }
        return liberalResult;
    }
    // If liberal parsing was not allowed or it failed, too, let's return the
    // (possibly failed) strict parsing result
    return strictResult;
}
exports.parseSpdxExpressionWithDetails = parseSpdxExpressionWithDetails;
var correctDashSeparatedLicenseIds = function (expression, upgradeGPLVariants) {
    var recurse = function (expression) {
        var fixLicenseRef = function (node) { return node; };
        var fixConjunctionInfo = function (node) {
            return { conjunction: node.conjunction, left: recurse(node.left), right: recurse(node.right) };
        };
        if (expression.conjunction) {
            return fixConjunctionInfo(expression);
        }
        else if (expression.licenseRef) {
            return fixLicenseRef(expression);
        }
        else {
            return (0, licenses_1.fixDashedLicenseInfo)(expression, upgradeGPLVariants);
        }
    };
    return recurse(expression);
};
